import json

from predeploy.predeploy import PreDeploymentValidation
from common.utils import run_command
from common.objects import ProcessOutput, Level, ProcessOutputs


class CfnLint(PreDeploymentValidation):

    def __init__(self, logger, test_case, template_path):
        super().__init__(logger, test_case, template_path)

    def sub_process_name(self) -> str:
        return "Cfn-Lint"

    """
    Run the cfn linting for the template provided and produces errors and warnings.
    """

    def sub_process(self) -> ProcessOutputs:
        process_outputs = ProcessOutputs()
        try:
            commands = ["cfn-lint", "-t", self.template_path, "-f", "json"]
            if self.logger.level == "DEBUG":
                commands.extend(["-D", "True"])
            elif self.logger.level == "INFO":
                commands.extend(["-I", "True"])
            self.logger.debug("Running command for CFN Lint as %s", ' '.join(commands))

            response = run_command(commands)
            self.logger.debug(
                "CFN Lint for the template %s is complete with result as %s." % (self.template_path, response))

            self.transform_process_output_data(json.loads(response), process_outputs)
        except Exception as exception:
            error = ProcessOutput(level=Level.ERROR, resource_name="CFN Lint Exception", message=str(exception))
            process_outputs.append(error)
            self.logger.error(
                "Test Case -> %s, CFN Lint Failed with %s." % (self.test_name, str(process_outputs)))
        return process_outputs

    def transform_process_output_data(self, issues, process_outputs):
        if issues:
            for issue in issues:
                error = ProcessOutput(level=Level.WARNING if "Warning" in issue["Level"] else Level.ERROR,
                                      line_number=issue["Location"]["Start"], resource_name="CFN Lint Validation Issue",
                                      message=issue["Rule"]["Id"] + " - " + issue["Message"])
                process_outputs.append(error)



from postdeploy.postdeploy import PostDeploymentValidation

from common.objects import ProcessOutputs, ProcessOutput, Level


class ParameterCheck(PostDeploymentValidation):

    def __init__(self, logger, test_case, template_path, stack, assertion):
        super().__init__(logger, test_case, template_path)
        self.stack = stack
        self.assertion = assertion

    def sub_process_name(self) -> str:
        return "Parameter-Check"

    """
        Validate the Parameters passed to Nested Stacks
    """

    def sub_process(self) -> ProcessOutputs:
        process_outputs = ProcessOutputs()
        # Get all Resources for Parent and Its Nested Stacks.
        all_parameters = self._get_all_parameters(self.stack, None)

        for resource_name, parameters in self.assertion.items():
            created_parameters = all_parameters.get(resource_name, None)
            failed_parameter = {}
            if created_parameters:
                for parameter_name, parameter_value in parameters.items():
                    created_value = self._get_created_value(parameter_name, created_parameters)
                    if parameter_value not in created_value:
                        failed_parameter[parameter_name] = "Parameter check failed with Expected Value -> " \
                                                           + parameter_value + \
                                                           " and Created Value -> " + created_value
            else:
                error = ProcessOutput(level=Level.ERROR,
                                      resource_name=resource_name,
                                      message="Expected resource is not created in the Stack.")
                process_outputs.append(error)
            if failed_parameter:
                error = ProcessOutput(level=Level.ERROR,
                                      resource_name=resource_name,
                                      message=failed_parameter)
                process_outputs.append(error)
        return process_outputs

    @staticmethod
    def _get_created_value(parameter_name, created_parameters):
        for value in created_parameters:
            if parameter_name == value["ParameterKey"]:
                return value["ParameterValue"]

    # Get all Nested Stack Physical Resource ID
    def _get_all_parameters(self, current_stack, logical_resource_id):
        all_parameters = {}
        if current_stack:
            for child in current_stack.children:
                child_resource = current_stack.resources.filter({"physical_id": child.stack_id})
                key = logical_resource_id + "." + child_resource[0].logical_id if logical_resource_id \
                    else child_resource[0].logical_id
                all_parameters[key] = child.created_parameters
                all_parameters.update(self._get_all_parameters(child, key))
        return all_parameters



import traceback
from enum import Enum, unique

from common.objects import ProcessOutput, Level, ProcessOutputs
from process import BaseProcess
from common.utils import get_instance_of_process


@unique
class AssertionTypes(Enum):
    ResourceExistence = "postdeploy.resource_validation.ResourceValidator"
    OutputsCheck = "postdeploy.output_validation.OutputValidator"
    ParameterCheck = "postdeploy.parameter_check.ParameterCheck"

    @classmethod
    def has_key(cls, name):
        return any(x for x in cls if x.name == name)

    @classmethod
    def all_keys(cls):
        return ", ".join(c.name for c in sorted(AssertionTypes))


class PostDeploymentValidation(BaseProcess):

    def __init__(self, logger, test_case, template_path, aws_client=None):
        super().__init__(logger, "Post-Deploy-Validations")
        self.test_case = test_case
        self.test_name = test_case["TestName"]
        self.region = test_case["Region"]
        self.test_data = test_case["TestData"]
        self.template_path = template_path
        self.parameters = self.test_data["parameters"]
        self.assertions = self.test_data["assertions"]

    # This method should be used by child classes to provide a human readable sub process name.
    def sub_process_name(self) -> str:
        raise NotImplementedError()

    # this method should be used to implement the sub_process process. The method should return ProcessOutputs object.
    def sub_process(self) -> ProcessOutputs:
        raise NotImplementedError()

    def process(self) -> dict:
        self.logger.info("Started stage -> %s for Test Name -> %s in Region -> %s.", self.process_name,
                         self.test_name, self.region)
        errors = {}
        current_stack = self.get_stack(self.region, self.test_name)
        if current_stack:
            for assertion in self.assertions:
                validation_errors = ProcessOutputs()
                sub_process_name = "Post Deploy"
                try:
                    class_instance = get_instance_of_process(
                        AssertionTypes[assertion["AssertType"]].value)(self.logger, self.test_case, self.template_path,
                                                                       current_stack, assertion["Assert"])
                    sub_process_name = class_instance.sub_process_name()
                    validation_errors = class_instance.sub_process()
                except Exception as exception:
                    # traceback.print_exc()
                    error = ProcessOutput(level=Level.ERROR, resource_name="Post Deploy Exception",
                                          message=str(exception))
                    validation_errors.append(error)
                finally:
                    if validation_errors:
                        errors[sub_process_name] = validation_errors
        else:
            self.logger.info("No Stack Found with Test Name -> %s and Region -> %s.", self.test_name, self.region)
        self.logger.info("Completed stage -> %s for Test Name -> %s in Region -> %s.", self.process_name,
                         self.test_name, self.region)
        return errors



import json

from predeploy.predeploy import PreDeploymentValidation
from common.utils import run_command
from common.objects import ProcessOutput, Level, ProcessOutputs


class CfnNag(PreDeploymentValidation):

    def __init__(self, logger, test_case, template_path):
        super().__init__(logger, test_case, template_path)

    def sub_process_name(self) -> str:
        return "Cfn-Nag"

    """
    Run the cfn Nag for the template provided and produces errors and warnings.
    """

    def sub_process(self) -> ProcessOutputs:
        process_outputs = ProcessOutputs()
        try:
            commands = ["cfn_nag", self.template_path, "-u", "json"]
            if self.logger.level == "DEBUG":
                commands.extend(["-d", "True"])
            self.logger.debug("Running command for CFN Nag as %s", ' '.join(commands))

            response = run_command(commands)
            self.logger.debug(
                "CFN Nag for the template %s is complete with result as %s." % (self.template_path, response))

            self.transform_process_output_data(json.loads(response), process_outputs)
        except Exception as exception:
            error = ProcessOutput(level=Level.ERROR, resource_name="CFN Nag Exception", message=str(exception))
            process_outputs.append(error)
            self.logger.error(
                "Test Case -> %s, CFN Nag Failed with %s." % (self.test_name, str(process_outputs)))
        return process_outputs

    @staticmethod
    def transform_process_output_data(issues, process_outputs):
        if issues:
            for issue in issues:
                for validation in issue["file_results"]["violations"]:
                    error = ProcessOutput(level=Level.WARNING if "WARN" in validation["type"] else Level.ERROR,
                                          line_number=CfnNag.transform_line_numbers(validation["line_numbers"]),
                                          resource_name=validation["logical_resource_ids"],
                                          message=validation["id"] + " - " + validation["message"])
                    process_outputs.append(error)

    @staticmethod
    def transform_line_numbers(line_numbers):
        output = []
        for line_number in line_numbers:
            output.append({'LineNumber': line_number})
        return output



import json

import cfn_flip

from common.utils import read_file
from common.objects import ProcessOutput, Level, ProcessOutputs

from deploy.deploy import Deploy


class ParameterCheck(Deploy):

    def __init__(self, logger, test_case, template_path, aws_client):
        super().__init__(logger, test_case, template_path, aws_client)
        self.parameters = self.test_data["parameters"]
        self.assertions = self.test_data["assertions"]

    def sub_process_name(self) -> str:
        return "Parameters-Check"

    """
    Run the Parameter check for the template provided and produces errors and warnings.
    """

    def sub_process(self) -> ProcessOutputs:
        process_outputs = ProcessOutputs()
        # Always convert to JSON irrespective of input as JSON or YAML.
        output_file = cfn_flip.to_json(read_file(self.template_path))
        json_data = json.loads(output_file)
        if "Parameters" in json_data:
            all_parameters = json_data["Parameters"]
            for parameter_name, parameter_value in self.parameters.items():
                if parameter_name not in all_parameters:
                    error = ProcessOutput(level=Level.ERROR, resource_name="Parameter Validation Issue",
                                          message="Provided Test Parameter -> " + parameter_name +
                                                  " is not present in the CloudFormation Template parameters.")
                    process_outputs.append(error)
        return process_outputs



import datetime
import re
import time
import traceback
from datetime import datetime
from typing import List

from common.logger import get_logger
from tabulate import tabulate
from common.objects import Events, Resources, Stacks, Event, Resource, Outputs, Output


class StackStatus:
    COMPLETE = ["CREATE_COMPLETE", "UPDATE_COMPLETE", "DELETE_COMPLETE"]
    IN_PROGRESS = [
        "CREATE_IN_PROGRESS",
        "DELETE_IN_PROGRESS",
        "UPDATE_IN_PROGRESS",
        "UPDATE_COMPLETE_CLEANUP_IN_PROGRESS",
        "ROLLBACK_IN_PROGRESS"
    ]
    FAILED = [
        "DELETE_FAILED",
        "CREATE_FAILED",
        "ROLLBACK_IN_PROGRESS",
        "ROLLBACK_FAILED",
        "ROLLBACK_COMPLETE",
        "UPDATE_ROLLBACK_IN_PROGRESS",
        "UPDATE_ROLLBACK_FAILED",
        "UPDATE_ROLLBACK_COMPLETE_CLEANUP_IN_PROGRESS",
        "UPDATE_ROLLBACK_COMPLETE",
    ]
    CREATE_FAIL_REASON = [
        "CREATE_FAILED",
        "ROLLBACK_IN_PROGRESS"
    ]


class Capabilities:
    IAM = "CAPABILITY_IAM"
    NAMED_IAM = "CAPABILITY_NAMED_IAM"
    AUTO_EXPAND = "CAPABILITY_AUTO_EXPAND"
    ALL = [IAM, NAMED_IAM, AUTO_EXPAND]


class IterStack(type):
    def __iter__(cls):
        return iter(cls._allStacks)


class Stack(metaclass=IterStack):
    _allStacks = []

    def __init__(self, test_name, region, test_data, template_path, logger, aws_client, stack_name=None, stack_id=None):
        self.logger = logger if logger else get_logger(__name__)

        # Test Details
        self.test_name = test_name
        self.template_path = template_path
        self.test_data = test_data
        self.parameters = test_data["parameters"]
        self.region = region

        # Stack Details
        self.parameters = self._generate_parameters(self.parameters)
        self.stack_name = stack_name if stack_name else re.sub('[^A-Za-z0-9]+', '', test_name + "_" + self.region)
        self.aws_client = aws_client
        self.cf_client_object = self.aws_client.cf_client_instance(self.region)

        # Stack Resources after Stack creation is done
        self.events = Events()
        self.resources = Resources()
        self.children = Stacks()
        self.outputs = Outputs()

        # Additional Stack Properties
        self.stack_id = stack_id
        self.change_set_id: str = ""
        self.created_parameters: List[dict] = []
        self.creation_time: datetime = datetime.fromtimestamp(0)
        self.status: str = ""
        self.status_reason: str = ""
        self.capabilities: List[str] = []
        self.tags: List[dict] = []
        self.parent_id: str = ""
        self.root_id: str = ""

    def deploy_template(self):
        deployment_error = []
        # Add Stacks to iter object when you are creating the Stack.
        self._allStacks.append(self)
        ll_stack_exists = False
        try:
            # -- Check if this stack name already exists
            existing_stack = self.aws_client.describe_stacks(self.cf_client_object, self.stack_name)
            if existing_stack:
                ll_stack_exists = True
                self.stack_id = existing_stack["StackId"]
        except Exception as e:
            ll_stack_exists = False

        try:
            # -- If the stack already exists then delete it first
            if ll_stack_exists:
                self.logger.info("As Stack %s Already Exist, Deleting the Stack First." % self.stack_name)
                self.delete_stack()

            # -- Create Stack
            response = self.aws_client.create_stack(self.cf_client_object, self.stack_name, self.template_path,
                                                    self.parameters, Capabilities.ALL)
            self.stack_id = response["StackId"]
            current_status = self._check_status()
            self.refresh()
            self._print_stack_events(self.events)
            if current_status in StackStatus.FAILED:
                self.logger.info("Stack %s Creation Failed with Status as %s." % (self.stack_name, current_status))
                status = "FAIL"
                deployment_error = [{"Resources": self.stack_name, "Message": self._get_failure_reasons()}]
            else:
                self.logger.info("Stack %s Creation Complete." % self.stack_name)
                status = "PASS"
        except Exception as e:
            # traceback.print_exc()
            status = "FAIL"
            deployment_error = [{"Resources": self.stack_name, "Message": str(e)}]
        return status, deployment_error

    def _get_failure_reasons(self):
        deployment_error = []
        for event in self.events:
            if event.status in StackStatus.CREATE_FAIL_REASON:
                deployment_error.append({"FailedStackName": self.stack_name,
                                         "Resource": event.logical_id,
                                         "Message": event.status_reason, })
                children = self.children.filter({"stack_id": event.physical_id})
                if children:
                    deployment_error.extend(children[0]._get_failure_reasons())
        return deployment_error

    def refresh(self, properties: bool = True, events: bool = True, resources: bool = True, children: bool = True):
        if properties:
            self._set_stack_properties()
        if events:
            self._fetch_events()
        if resources:
            self._fetch_resources()
        if children:
            self._fetch_children()

    @staticmethod
    def _generate_parameters(parameters):
        stack_parameters = []
        if parameters:
            for parameter_name, parameter_value in parameters.items():
                stack_parameters.append({"ParameterKey": parameter_name, "ParameterValue": parameter_value})
        return stack_parameters

    def _check_status(self):
        current_stack = self.aws_client.describe_stacks(self.cf_client_object, self.stack_id)
        current_status = None
        if current_stack:
            current_status = current_stack["StackStatus"]
            for ln_loop in range(1, 9999):
                self.logger.debug("Inside the loop with count as %s" % ln_loop)
                if current_status in StackStatus.IN_PROGRESS:
                    time.sleep(30)
                    try:
                        current_stack = self.aws_client.describe_stacks(self.cf_client_object, self.stack_id)
                    except Exception as e:
                        current_status = "DELETE_COMPLETE"
                        break
                    if current_stack["StackStatus"] != current_status:
                        current_status = current_stack["StackStatus"]
                else:
                    break
                self.logger.info("Stack -> %s, Current Status -> %s.", self.stack_name, current_status)
        return current_status

    def delete_stack(self):
        current_status = self._check_status()
        self.logger.debug("Starting Stack %s deletion with current status as %s" % (self.stack_name, current_status))
        for i in range(1, 10):
            if current_status == "DELETE_FAILED":
                self.refresh(properties=False, events=False, children=False)
                failed_resources = self.resources.filter({"status": "DELETE_FAILED"})
                failed_resources_ids = [failed_resource.logical_id for failed_resource in failed_resources]
                self.logger.info("Retain Resources %s and delete the stack %s again." % (
                    ', '.join(failed_resources_ids), self.stack_name))
                self.aws_client.delete_stack(self.cf_client_object, self.stack_id, failed_resources_ids)
            elif current_status not in StackStatus.IN_PROGRESS:
                child_stack_ids = self._get_child_stack_ids()
                for child_stack_id in child_stack_ids:
                    self.aws_client.delete_stack(self.cf_client_object, child_stack_id)
            current_status = self._check_status()

            if current_status != "DELETE_COMPLETE":
                continue
            else:
                self.logger.info("Complete Stack %s deletion with current status as DELETE_COMPLETE" % self.stack_name)
                break
        return current_status

    def _get_child_stack_ids(self):
        stack_ids = set()
        stack_ids.add(self.stack_id)
        for child_stack in self.children:
            stack_ids.update(child_stack._get_child_stack_ids())
        return stack_ids

    def _set_stack_properties(self):
        current_stack = self.aws_client.describe_stacks(self.cf_client_object, self.stack_id)
        if current_stack:
            self.change_set_id = current_stack["ChangeSetId"] if "ChangeSetId" in current_stack else None
            self.created_parameters = current_stack["Parameters"] if "Parameters" in current_stack else None
            self.creation_time = current_stack["CreationTime"] if "CreationTime" in current_stack else None
            self.status = current_stack["StackStatus"] if "StackStatus" in current_stack else None
            self.status_reason = current_stack["StackStatusReason"] if "StackStatusReason" in current_stack else None
            self.capabilities = current_stack["Capabilities"] if "Capabilities" in current_stack else None
            if "Outputs" in current_stack:
                for output in current_stack["Outputs"]:
                    self.outputs.append(Output(output))
            self.tags = current_stack["Tags"] if "Tags" in current_stack else None
            self.parent_id = current_stack["ParentId"] if "ParentId" in current_stack else None
            self.root_id = current_stack["RootId"] if "RootId" in current_stack else None
            self.logger.debug("Successfully Fetched Stack Properties " + self.stack_name)

    def _fetch_resources(self):
        resources = Resources()
        response = self.aws_client.list_stack_resources(self.cf_client_object, self.stack_id)
        if response:
            for resource in response:
                resources.append(Resource(resource))
        self.resources = resources
        self.logger.debug("Successfully Fetched Stack Resources " + self.stack_name)

    def _fetch_events(self):
        events = Events()
        response = self.aws_client.list_stack_events(self.cf_client_object, self.stack_id)
        if response:
            for event in response:
                events.append(Event(event))
        self.events = events
        self.logger.debug("Successfully Fetched Stack Events " + self.stack_name)

    # Add Children Stack Properties, Outputs, Resources
    def _fetch_children(self):
        for resource in self.resources:
            if resource.physical_id and "arn:aws:cloudformation" in resource.physical_id \
                    and "WaitCondition" not in resource.type:
                child_stack = self.aws_client.describe_stacks(self.cf_client_object, resource.physical_id)
                if child_stack:
                    stack_obj = Stack(test_name=self.test_name, region=self.region, test_data=self.test_data,
                                      template_path=self.template_path, logger=self.logger,
                                      stack_name=child_stack["StackName"], stack_id=child_stack["StackId"],
                                      aws_client=self.aws_client)
                    stack_obj.refresh()
                    self.children.append(stack_obj)
        self.logger.debug("Successfully Fetched Stack Children " + self.stack_name)

    def _print_stack_events(self, events):
        self.logger.info(
            "********** Events -> Stack Name -> %s, Region -> %s **********" % (self.stack_name, self.region))
        headers = ['Timestamp', 'Resource Name', 'Resource Type', 'Resource Status', 'Resource Status Reason']
        rows = []
        for event in events:
            rows.append([event.timestamp, event.logical_id, event.type, event.status, event.status_reason])

        self.logger.info("\n" + tabulate(rows, headers=headers, tablefmt='orgtbl') + "\n")

        self.logger.info("************************** Events End **************************")



import traceback
import os
import sys

cur_dir = os.path.abspath(os.path.dirname(__file__))
sys.path.insert(0, cur_dir)


def _import_all_modules():
    """ Dynamically imports all modules in this package. """

    import os
    global __all__
    __all__ = []
    globals_, locals_ = globals(), locals()

    # Dynamically import all the package modules in this file's directory.
    for filename in os.listdir(os.path.dirname(__file__)):
        # Process all python files in directory that don't start
        # with underscore (which also prevents this module from
        # importing itself).
        if filename[0] != '_' and filename.split('.')[-1] in ('py', 'pyw'):
            module_name = filename.split('.')[0]  # Filename sans extension.
            package_module = '.'.join([__name__, module_name])
            try:
                module = __import__(package_module, globals_, locals_, [module_name])
            except:
                traceback.print_exc()
                raise
            for name in module.__dict__:
                if not name.startswith('_'):
                    globals_[name] = module.__dict__[name]
                    __all__.append(name)


_import_all_modules()



from postdeploy.postdeploy import PostDeploymentValidation

from common.objects import ProcessOutputs, ProcessOutput, Level


class ResourceValidator(PostDeploymentValidation):

    def __init__(self, logger, test_case, template_path, stack, assertion):
        super().__init__(logger, test_case, template_path)
        self.stack = stack
        self.assertion = assertion

    def sub_process_name(self) -> str:
        return "Resource-Validation"

    """
    Validate the resources
    """

    def sub_process(self) -> ProcessOutputs:
        process_outputs = ProcessOutputs()
        if "Resources" in self.assertion:
            # Get all Resources for Parent and Its Nested Stacks.
            all_created_resources = self._get_resource_logical_id(self.stack, None)

            assert_resources = self.assertion["Resources"]
            failed_resources = []
            if isinstance(assert_resources, list):
                for assert_resource in assert_resources:
                    if assert_resource not in all_created_resources:
                        failed_resources.append(assert_resource)
                    else:
                        all_created_resources.remove(assert_resource)
                if all_created_resources:
                    error = ProcessOutput(level=Level.WARNING,
                                          resource_name=all_created_resources,
                                          message="Extra resource created in the Stack.")
                    process_outputs.append(error)
                if failed_resources:
                    error = ProcessOutput(level=Level.ERROR,
                                          resource_name=failed_resources,
                                          message="Expected resource is not created in the Stack.")
                    process_outputs.append(error)
            else:
                if len(all_created_resources) > 0:
                    error = ProcessOutput(level=Level.WARNING,
                                          resource_name=all_created_resources,
                                          message="Extra resource created in the Stack.")
                    process_outputs.append(error)
        return process_outputs

    # Get all Nested Stack Physical Resource ID
    def _get_resource_logical_id(self, current_stack, logical_resource_id):
        all_created_resources = []
        if current_stack.resources:
            for resource in current_stack.resources:
                resource_id = logical_resource_id + "." + resource.logical_id if logical_resource_id \
                    else resource.logical_id
                all_created_resources.append(resource_id)
                nested_stack = current_stack.children.filter({"stack_id": resource.physical_id})
                if nested_stack:
                    all_created_resources.extend(self._get_resource_logical_id(nested_stack[0], resource_id))
        return all_created_resources



from process import BaseProcess

from common.objects import ProcessOutputs, ProcessOutput, Level


class PreDeploymentValidation(BaseProcess):
    """
        Aim of this file is to process the test file.
        - If test file passes schema checks, file checks
        - Go for CFN-lint and CFN-Nag validation on the CF template.
        - More classes can be added by implementing PreDeploymentValidation class.
            - Each child class should implement sub_process_name and validate function.
    """

    def __init__(self, logger, test_case, template_path, aws_client=None):
        super().__init__(logger, "Pre-Deploy-Validations")
        self.test_case = test_case
        self.test_name = test_case["TestName"]
        self.region = test_case["Region"]
        self.test_data = test_case["TestData"]
        self.template_path = template_path

    # This method should be used by child classes to provide a human readable sub process name.
    def sub_process_name(self) -> str:
        raise NotImplementedError()

    # this method should be used to implement the sub_process process. The method should return ProcessOutputs object.
    def sub_process(self) -> ProcessOutputs:
        raise NotImplementedError()

    def process(self) -> dict:
        self.logger.info("Started stage -> %s for Test Name -> %s in Region -> %s.", self.process_name, self.test_name,
                         self.region)
        errors = {}
        for sub_class in PreDeploymentValidation.__subclasses__():
            validation_errors = ProcessOutputs()
            sub_process_name = "Pre Deploy"
            try:
                class_instance = sub_class(self.logger, self.test_case, self.template_path)
                sub_process_name = class_instance.sub_process_name()
                validation_errors = class_instance.sub_process()
            except Exception as exception:
                error = ProcessOutput(level=Level.ERROR, resource_name="Pre Deploy Exception", message=str(exception))
                validation_errors.append(error)
            finally:
                if validation_errors:
                    errors[sub_process_name] = validation_errors
        self.logger.info("Completed stage -> %s for Test Name -> %s in Region -> %s.", self.process_name,
                         self.test_name,
                         self.region)
        return errors



import argparse

from process import ProcessFlow
from common.logger import get_logger


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-f", '--test-file', dest='test_file', required=True, help='Provide the test file.')
    parser.add_argument("-d", '--debug-logging', dest='debug_logging', help='Provide to enable Debug Logging.')

    args = parser.parse_args()

    # Enable Logger
    logger = get_logger(__name__, "DEBUG" if args.debug_logging else "INFO")

    # Pass it to the Pre validation file for schema check, linting check, Security check.
    # If everything passes on pre validation level, pass it for deployment.
    # If deployment is successful, pass it for Post deployment validation
    ProcessFlow(args.test_file, logger)


if __name__ == '__main__':
    main()



import json
from datetime import datetime
from enum import Enum
from uuid import UUID, uuid4


def criteria_matches(criteria: dict, instance):
    for k in criteria:
        if k not in instance.__dict__:
            raise ValueError("%s is not a valid property %s" % (k, type(instance)))
    for k, v in criteria.items():
        if getattr(instance, k) != v:
            return False
        elif k == "Timestamp" and getattr(instance, k) < v:
            return False
    return True


class FilterableList(list):
    def filter(self, criteria):
        if not criteria:
            return self
        f_list = FilterableList()
        for item in self:
            if criteria_matches(criteria, item):
                f_list.append(item)
        return f_list


class Level(str, Enum):
    ERROR = 'ERROR'
    WARNING = 'WARNING'


class ProcessOutput:
    """
    A ProcessOutput object with TimeStamp, Level, Location abd Message as compulsory data.
    """

    def __init__(self, level, message, line_number=None, resource_name=None, test_name=None,
                 region=None, stack_name=None):
        self.Timestamp: datetime = datetime.now()
        self.Level: Level = level
        if line_number:
            self.Location: str = line_number
        self.Message = message
        if test_name:
            self.TestName: str = test_name
        if stack_name:
            self.StackName: str = stack_name
        if region:
            self.Region = region
        if resource_name:
            self.Resource = resource_name

    def __str__(self):
        return "{} {} {}".format(self.Timestamp, self.Level, self.Message)

    def __repr__(self):
        return "<Event object {} at {}>".format(self.Message, hex(id(self)))


class ProcessOutputs(FilterableList):
    pass


class Stacks(FilterableList):
    pass


class Resource:
    def __init__(self, resource_dict: dict):
        self.logical_id: str = resource_dict["LogicalResourceId"]
        self.type: str = resource_dict["ResourceType"]
        self.status: str = resource_dict["ResourceStatus"]
        self.physical_id: str = ""
        self.last_updated_timestamp: datetime = datetime.fromtimestamp(0)
        self.status_reason: str = ""
        if "PhysicalResourceId" in resource_dict.keys():
            self.physical_id = resource_dict["PhysicalResourceId"]
        if "LastUpdatedTimestamp" in resource_dict.keys():
            self.last_updated_timestamp = resource_dict["LastUpdatedTimestamp"]
        if "ResourceStatusReason" in resource_dict.keys():
            self.status_reason = resource_dict["ResourceStatusReason"]

    def __str__(self):
        return "<Resource {} {}>".format(self.logical_id, self.status)


class Resources(FilterableList):
    pass


class Event:
    def __init__(self, event_dict: dict):
        self.event_id: str = event_dict["EventId"]
        self.stack_name: str = event_dict["StackName"]
        self.logical_id: str = event_dict["LogicalResourceId"]
        self.type: str = event_dict["ResourceType"]
        self.status: str = event_dict["ResourceStatus"]
        self.physical_id: str = ""
        self.timestamp: datetime = datetime.fromtimestamp(0)
        self.status_reason: str = ""
        self.properties: dict = {}
        if "PhysicalResourceId" in event_dict.keys():
            self.physical_id = event_dict["PhysicalResourceId"]
        if "Timestamp" in event_dict.keys():
            self.timestamp = event_dict["Timestamp"]
        if "ResourceStatusReason" in event_dict.keys():
            self.status_reason = event_dict["ResourceStatusReason"]
        if "ResourceProperties" in event_dict.keys():
            self.properties = json.loads(event_dict["ResourceProperties"])

    def __str__(self):
        return "{} {} {}".format(self.timestamp, self.logical_id, self.status)

    def __repr__(self):
        return "<Event object {} at {}>".format(self.event_id, hex(id(self)))


class Events(FilterableList):
    pass


class Output:
    def __init__(self, output_dict: dict):
        self.key: str = output_dict["OutputKey"]
        self.value: str = output_dict["OutputValue"]
        self.description: str = ""
        self.export_name: str = ""
        if "Description" in output_dict.keys():
            self.description = output_dict["Description"]
        if "ExportName" in output_dict.keys():
            self.export_name = output_dict["ExportName"]


class Outputs(FilterableList):
    pass



from postdeploy.postdeploy import PostDeploymentValidation

from common.objects import ProcessOutputs, ProcessOutput, Level


class OutputValidator(PostDeploymentValidation):

    def __init__(self, logger, test_case, template_path, stack, assertion):
        super().__init__(logger, test_case, template_path)
        self.stack = stack
        self.assertion = assertion

    def sub_process_name(self) -> str:
        return "Output-Validation"

    """
        Validate the outputs
        """

    def sub_process(self) -> ProcessOutputs:
        process_outputs = ProcessOutputs()
        if "Outputs" in self.assertion:
            # Get all Resources for Parent and Its Nested Stacks.
            all_created_outputs = self._get_output_keys(self.stack, None)

            assert_outputs = self.assertion["Outputs"]
            if isinstance(assert_outputs, list):
                failed_outputs = []
                for assert_output in assert_outputs:
                    if assert_output not in all_created_outputs:
                        failed_outputs.append(assert_output)
                    else:
                        all_created_outputs.remove(assert_output)
                if all_created_outputs:
                    error = ProcessOutput(level=Level.WARNING,
                                          resource_name=all_created_outputs,
                                          message="Extra Outputs created in the Stack.")
                    process_outputs.append(error)
                if failed_outputs:
                    error = ProcessOutput(level=Level.ERROR,
                                          resource_name=failed_outputs,
                                          message="Expected output is not created in the Stack.")
                    process_outputs.append(error)
            else:
                if len(all_created_outputs) > 0:
                    error = ProcessOutput(level=Level.WARNING,
                                          resource_name=all_created_outputs,
                                          message="Extra resource created in the Stack.")
                    process_outputs.append(error)
        return process_outputs

    # Get all Nested Stack Physical Resource ID
    def _get_output_keys(self, current_stack, logical_resource_id):
        all_created_outputs = []
        if current_stack.outputs:
            for output in current_stack.outputs:
                output_key = logical_resource_id + "." + output.key if logical_resource_id else output.key
                all_created_outputs.append(output_key)

        for child in current_stack.children:
            child_resource = current_stack.resources.filter({"physical_id": child.stack_id})
            key = logical_resource_id + "." + child_resource[0].logical_id if logical_resource_id \
                else child_resource[0].logical_id
            all_created_outputs.extend(self._get_output_keys(child, key))
        return all_created_outputs



import datetime
import json
import os

from common.utils import load_json_file, get_absolute_path

from common.objects import Level


class ReportGeneration(object):

    # Create Report as per test case with all stages it passed through.
    @staticmethod
    def generate_report(output, test_file_path, template_path, logger):

        all_test_cases = {}
        skipped_cases = 0
        json_data = load_json_file(os.path.realpath(test_file_path))
        total_tests = 0
        for test in json_data["Tests"]:
            if "Skip" in test and test["Skip"]:
                skipped_cases = skipped_cases + 1
                total_tests += 1
            else:
                for region in test["Regions"]:
                    all_test_cases[test["TestName"] + "#" + region] = "PASS"
                    total_tests += 1

        test_cases = []
        if output:
            for element in output:
                processes = {}
                test_case_errors = 0
                process_names = element["Validations"]
                test_name = element["TestName"]
                region = element["Region"]
                element["Status"] = "PASS"
                for process in process_names:
                    for process_name, sub_process_names in process.items():
                        sub_processes = []
                        for sub_process_name, issues in sub_process_names.items():
                            error_warning = {}
                            if issues:
                                errors = len(issues.filter({"Level": Level.ERROR}))
                                warnings = len(issues.filter({"Level": Level.WARNING}))
                                if errors > 0 or warnings > 0:
                                    error_warning[sub_process_name] = {"Errors": errors, "Warnings": warnings}
                                test_case_errors += errors
                            if error_warning:
                                sub_processes.append(error_warning)
                        if sub_processes:
                            processes[process_name] = sub_processes
                if processes:
                    test_cases.append(
                        {"TestName": test_name, "Region": region, "Status": "FAIL" if test_case_errors > 0 else "PASS",
                         "Validations": processes})
                if test_case_errors > 0:
                    element["Status"] = "FAIL"
                    all_test_cases[test_name + "#" + region] = "FAIL"

        failed_cases = sum(status == "FAIL" for status in all_test_cases.values())
        passed_cases = sum(status == "PASS" for status in all_test_cases.values())

        if (passed_cases + failed_cases + skipped_cases) != total_tests:
            raise Exception(
                "Issue with Report Generation as Total Test cases(%s) "
                "does not match Passed(%s) + Failed(%s)+ Skipped(%s) Test Cases" % (
                    total_tests, passed_cases, failed_cases, skipped_cases))

        # Create Result file
        with open("report.json", 'w') as parameter:
            json.dump(output, parameter, indent=4, default=ReportGeneration.object_to_dictionary)

        # Creating report data
        report = {"Template": template_path,
                  "ReportPath": os.path.abspath("report.json"),
                  "Status": "FAIL" if failed_cases > 0 else "PASS",
                  "TestResult": {"Total": total_tests, "Pass": passed_cases, "Fail": failed_cases,
                                 "Skipped": skipped_cases}}
        # "Report_Status": test_cases}

        logger.info("Report for CloudFormation Testing is : \n %s" % (json.dumps(report, indent=4)))

    @staticmethod
    def object_to_dictionary(obj):
        if isinstance(obj, datetime.date):
            return obj.strftime("%m/%d/%Y, %H:%M:%S %p")
        else:
            return obj.__dict__



import os
import traceback

import jsonschema

from common.utils import load_json_file, get_absolute_path, check_valid_file
from common.objects import ProcessOutput, Level, ProcessOutputs

from postdeploy.postdeploy import AssertionTypes

from process import BaseProcess


class TestFileValidator(BaseProcess):
    RESOURCE_FOLDER = os.path.join(os.path.dirname(os.path.realpath(__file__)), "resources")

    def __init__(self, logger, test_file_name):
        super().__init__(logger, test_file_name)
        # Loading Schema Files
        self.test_schema = load_json_file(os.path.join(self.RESOURCE_FOLDER, "test_file.schema"))
        self.property_schema = load_json_file(os.path.join(self.RESOURCE_FOLDER, "parameter_file.schema"))
        # Get Full Path of test file
        self.test_file_full_path = os.path.realpath(test_file_name)

    """
    Read and validate the test JSON file. Validate Template and parameters JSON.
    Also, replace relative path with absolute path in the JSON object itself.
    """

    def process(self) -> ProcessOutputs:
        process_outputs = ProcessOutputs()
        try:
            json_data = load_json_file(self.test_file_full_path)
            self.logger.debug("Reading of the file -> %s is complete.", self.test_file_full_path)
            if self.validate_json_using_schema(json_data, self.test_schema):
                success = self.validate_paths_in_test_file(json_data)
                if success:
                    if self.validate_duplicates(json_data):
                        self.logger.debug(
                            "Provided File -> %s passed all validation checks." % self.test_file_full_path)
                        return process_outputs
        except Exception as exception:
            # traceback.print_exc()
            error = ProcessOutput(level=Level.ERROR, resource_name="Test File Validation Issue", message=str(exception))
            process_outputs.append(error)
            self.logger.error(
                "Provided File -> %s failed validation checks %s." % (self.test_file_full_path, str(process_outputs)))
        return process_outputs

    def validate_json_using_schema(self, json_data, json_schema):
        self.logger.debug("Validating the file with data as %s." % str(json_data))
        try:
            jsonschema.validate(json_data, json_schema)
            self.logger.debug("Validation complete for the file.")
            return True
        except jsonschema.ValidationError as e:
            raise e

    def validate_paths_in_test_file(self, json_data):
        # Check if the template path is valid and is File.
        template_path = get_absolute_path(json_data.get("Global").get("TemplatePath"), self.test_file_full_path)
        if check_valid_file(template_path):
            self.logger.debug("Provided Template path -> %s" % template_path)
            # Check all parameters files in the test file.
            if "Tests" in json_data:
                for test in json_data.get("Tests"):
                    if "Parameters" in test:
                        parameters = test.get("Parameters")
                        if "Path" in parameters:
                            parameter_path = get_absolute_path(parameters["Path"], self.test_file_full_path)
                            if check_valid_file(parameter_path):
                                self.validate_json_using_schema(load_json_file(parameter_path), self.property_schema)
                                self.logger.debug("Provided Parameter path is %s" % parameter_path)
        return True

    @staticmethod
    def validate_duplicates(json_data):
        test_names = []
        if "Tests" in json_data:
            for test in json_data.get("Tests"):
                test_name = test["TestName"]
                assertions = test["Assertions"]
                if test_name in test_names:
                    raise Exception("Test Name - %s is duplicate in the provided Test File. "
                                    "Please make sure Test Names are unique." % test_name)
                else:
                    test_names.append(test_name)
                for assertion in assertions:
                    if not AssertionTypes.has_key(assertion["AssertType"]):
                        raise Exception("Test Name - %s has Assertion Type %s that is not valid. "
                                        "Please make sure Assertion Types is one of the following %s." %
                                        (test_name, assertion["AssertType"], AssertionTypes.all_keys()))
        return True



import logging
import sys

FORMATTER = logging.Formatter("%(levelname)s | %(asctime)s | %(threadName)s | %(filename)s | %(message)s")


def get_console_handler():
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(FORMATTER)
    return console_handler


def get_logger(logger_name, log_level="DEBUG"):
    name = logger_name or __name__
    logger = logging.getLogger(name)
    if not logger.handlers:
        logger.setLevel(logging.getLevelName(log_level))
        logger.addHandler(get_console_handler())
        # with this pattern, it's rarely necessary to propagate the error up to parent
        logger.propagate = False
    return logger



import os
import threading

import boto3
import cfn_flip
from boto3.s3.transfer import S3Transfer

from common.utils import read_file


class AwsClient(object):

    def __init__(self):
        self.client_lock = threading.RLock()

    __cf_clients = {}
    file_url = None

    def cf_client_instance(self, region):
        with self.client_lock:
            if not self.__cf_clients.get(region):
                self.__cf_clients[region] = boto3.client("cloudformation", region)
        return self.__cf_clients.get(region)

    def s3_client_instance(self, region):
        with self.client_lock:
            s3_client = boto3.client("s3", region)
        return s3_client

    def describe_stacks(self, cf_client, stack_identifier):
        stack_details = None
        with self.client_lock:
            if stack_identifier:
                stack_details = cf_client.describe_stacks(StackName=stack_identifier)["Stacks"][0]
        return stack_details

    def create_stack(self, cf_client, stack_name, template_path, parameters, capability):
        response = None
        with self.client_lock:
            if stack_name:
                if self.file_url:
                    response = cf_client.create_stack(StackName=stack_name, TemplateURL=self.file_url,
                                                      Parameters=parameters, Capabilities=capability)
                else:
                    output_file = cfn_flip.to_json(read_file(template_path))
                    response = cf_client.create_stack(StackName=stack_name, TemplateBody=output_file,
                                                      Parameters=parameters, Capabilities=capability)
        return response

    def delete_stack(self, cf_client, stack_identifier, failed_resources_ids=None):
        response = None
        with self.client_lock:
            if stack_identifier:
                if failed_resources_ids:
                    cf_client.delete_stack(StackName=stack_identifier, RetainResources=failed_resources_ids)
                else:
                    cf_client.delete_stack(StackName=stack_identifier)
        return response

    def list_stack_resources(self, cf_client, stack_identifier):
        response = []
        with self.client_lock:
            if stack_identifier:
                for page in cf_client.get_paginator("list_stack_resources").paginate(StackName=stack_identifier):
                    response.extend(page["StackResourceSummaries"])
        return response

    def list_stack_events(self, cf_client, stack_identifier):
        response = []
        with self.client_lock:
            if stack_identifier:
                for page in cf_client.get_paginator("describe_stack_events").paginate(StackName=stack_identifier):
                    response.extend(page["StackEvents"])
        return response

    def create_bucket(self, s3_client, bucket_name):
        with self.client_lock:
            response = s3_client.create_bucket(Bucket=bucket_name)
        return response['Location']

    def upload_to_bucket(self, s3_client, bucket_name, template_path):
        with self.client_lock:
            transfer = S3Transfer(s3_client)
            head, tail = os.path.split(template_path)
            transfer.upload_file(template_path, bucket_name, tail, extra_args={'ACL': 'public-read'})
            self.file_url = '%s/%s/%s' % (s3_client.meta.endpoint_url, bucket_name, tail)
        return self.file_url

    def delete_bucket(self, bucket_name):
        if bucket_name:
            with self.client_lock:
                bucket = boto3.resource('s3').Bucket(bucket_name)
                bucket.objects.all().delete()
                bucket.delete()



import traceback

from common.objects import ProcessOutput, Level, ProcessOutputs

from cleanup.cleanup import CleanUp


class DeleteStack(CleanUp):

    def __init__(self, logger, test_case, template_path, stack):
        super().__init__(logger, test_case, template_path)
        self.parameters = self.test_data["parameters"]
        self.assertions = self.test_data["assertions"]
        self.stack = stack

    def sub_process_name(self) -> str:
        return "Delete-Stack"

    """
    Delete Stacks provided and produces errors and warnings.
    """

    def sub_process(self) -> ProcessOutputs:
        process_outputs = ProcessOutputs()
        try:
            self.stack.delete_stack()
        except Exception as e:
            # traceback.print_exc()
            error = ProcessOutput(level=Level.ERROR, message=str(e), stack_name=self.stack.stack_name)
            process_outputs.append(error)
        self.logger.debug("Delete Stack complete with errors as %s", process_outputs)
        return process_outputs



import importlib
import json
import os
import subprocess
import sys
import traceback

import cfn_flip
from common.objects import Level


def get_normalized_path(path):
    return os.path.realpath(os.path.abspath(os.path.expanduser(path)))


def check_valid_file(file_path):
    if not os.path.isfile(file_path):
        # traceback.print_exc()
        raise Exception("Provided path %s is not a valid file path." % file_path)
    return True


def get_absolute_path(relative_path, compare_file_path):
    if not os.path.isabs(relative_path):
        absolute_path = os.path.join(os.path.dirname(compare_file_path), relative_path)
        absolute_path = os.path.abspath(absolute_path)
        return absolute_path
    return relative_path


def read_file(file_path):
    with open(file_path, "r") as f:
        return f.read().strip()


def load_json_file(file_full_path):
    """
    Load the given Test JSON file.
    """
    content = {}
    if check_valid_file(file_full_path):
        if ".schema" in file_full_path:
            with open(file_full_path, 'r') as fp:
                content = fp.read()
        else:
            content = cfn_flip.to_json(read_file(file_full_path))
    return json.loads(content)


def _run(command, input=None, check=False, **kwargs):
    if sys.version_info >= (3, 5):
        return subprocess.run(command, capture_output=True)
    if input is not None:
        if 'stdin' in kwargs:
            raise ValueError('stdin and input arguments may not both be used.')
        kwargs['stdin'] = subprocess.PIPE

    process = subprocess.Popen(command, **kwargs)
    try:
        stdout, stderr = process.communicate(input)
    except:
        process.kill()
        process.wait()
        raise
    retcode = process.poll()
    if check and retcode:
        raise subprocess.CalledProcessError(
            retcode, process.args, output=stdout, stderr=stderr)
    return retcode, stdout, stderr


def run_command(cmdargs):
    resp = _run(cmdargs)
    if len(resp.stderr.decode()) > 0:
        # traceback.print_exc()
        raise Exception("Error in run command %s cmd: %s" % (resp, cmdargs))
    return resp.stdout


def get_instance_of_process(process_name):
    modname, _, cls_name = process_name.rpartition('.')
    mod = importlib.import_module(modname)
    return getattr(mod, cls_name)


def generate_data_each_test_case(test_file_full_path):
    json_data = load_json_file(test_file_full_path)
    json_data = json.loads(os.path.expandvars(json.dumps(json_data)))
    processed_data = {}
    global_parameters = {}
    template_path = get_absolute_path(json_data["Global"]["TemplatePath"], test_file_full_path)
    if "GlobalParameters" in json_data["Global"]:
        global_parameters = json_data["Global"]["GlobalParameters"]
    for test in json_data["Tests"]:
        if "Skip" not in test or not test["Skip"]:
            all_parameters = global_parameters.copy()
            if "Parameters" in test:
                parameters = test["Parameters"]
                if "Path" in parameters:
                    parameter_values = load_json_file(get_absolute_path(parameters['Path'], test_file_full_path))
                    all_parameters.update(parameter_values)
                elif "Values" in parameters:
                    all_parameters.update(parameters['Values'])
            assertions = []
            if "Assertions" in test:
                assertions = test["Assertions"]

            regions = test["Regions"]

            processed_data[test["TestName"]] = {"parameters": all_parameters, "assertions": assertions,
                                                "regions": regions}

    # Divide test cases by regions also
    new_data = []
    for test_case, test_data in processed_data.items():
        for region in test_data["regions"]:
            new_data.append({"TestName": test_case, "Region": region,
                             "TestData": {"parameters": test_data["parameters"],
                                          "assertions": test_data["assertions"]}})

    return template_path, new_data, json_data["Global"]["ParallelTestsRun"]


def check_for_error(process_name, process_output):
    if process_output:
        sub_process_outputs = process_output[process_name]
        if sub_process_outputs:
            for sub_process_name, output in sub_process_outputs.items():
                if output:
                    errors = len(output.filter({"Level": Level.ERROR}))
                    if errors > 0:
                        return True
    return False



from setuptools import setup, find_packages
from os.path import join, dirname, abspath
import io

here = abspath(dirname(__file__))

with open(join(here, 'VERSION')) as VERSION_FILE:
    __versionstr__ = VERSION_FILE.read().strip()


with open(join(here, 'requirements.txt')) as REQUIREMENTS:
    INSTALL_REQUIRES = REQUIREMENTS.read().split('\n')


with io.open(join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


CONSOLE_SCRIPTS = [
    'sumocfntester=sumocfntester.console:main'
]

setup(
    name="sumologic-cfn-tester",
    version=__versionstr__,
    packages=find_packages(),
    install_requires=INSTALL_REQUIRES,
    # PyPI metadata
    author="SumoLogic",
    author_email="it@sumologic.com, apps-team@sumologic.com",
    description="Sumo Logic CloudFormation tester allows you to tests the CF for validation and deployments.",
    license="PSF",
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords="sumologic cloudformation testing validation aws lint",
    url="https://github.com/SumoLogic/sumologic-aws-cloudformation-test",
    zip_safe=True,
    include_package_data=True,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.7',
        'Operating System :: OS Independent'
    ],
    entry_points={
        'console_scripts': CONSOLE_SCRIPTS,
    }

)



import json
import os
import random
import string
import sys
import time
import traceback
from abc import ABC, abstractmethod
from concurrent import futures

import cfn_flip
from common.logger import get_logger
from common.utils import check_for_error, get_instance_of_process, generate_data_each_test_case, read_file

from common.report import ReportGeneration

from common.cf_utility import Stack

from common.aws_client import AwsClient


class BaseProcess(ABC):
    def __init__(self, logger, process_name):
        self.logger = logger if logger else get_logger(__name__)
        self.process_name = process_name

    @abstractmethod
    def process(self) -> dict:
        raise NotImplementedError()

    def transform_output(self, output):
        if output:
            return {self.process_name: output}
        return None

    def get_stack(self, region, test_name):
        for stack in Stack:
            if stack.region == region and stack.test_name == test_name:
                return stack
        return None


class ProcessFlow(object):
    process_flow = ['predeploy.predeploy.PreDeploymentValidation',
                    'deploy.deploy.Deploy',
                    'postdeploy.postdeploy.PostDeploymentValidation',
                    ]

    def __init__(self, test_file, logger):
        super().__init__()
        self.test_file_name: str = test_file
        self.logger = logger if logger else get_logger(__name__)
        # Start the CloudFormation Testing Process Flow
        self.process()

    """
        Process flow iss
        1. Get the Class Instance as per the Flow. PreDeploy, Deploy, PostDeploy and CleanUp.
        2. Get the Output for each Process
        3. Generate Report for each process
        4. Append the output of each process to full output
        5. If errors in any stage, raise exception, stop process
        6. Print report and generate report.json file
    """

    def process(self):
        # First Validate the Test JSON File. If that Passes, all validation should be performed on all Test Cases.
        validator_class = get_instance_of_process("common.test_file_validator.TestFileValidator")(self.logger,
                                                                                                  self.test_file_name)
        test_file_validation_errors = validator_class.process()
        if not test_file_validation_errors:
            # Get test case for with all parameters (union of global and test case parameters).
            # Test case will be test case * regions provided in each test case
            template_path, test_data, max_workers = generate_data_each_test_case(os.path.realpath(self.test_file_name))
            all_futures = {}
            full_output = []
            aws_client = AwsClient()
            # Upload template to S3, if size exceed 51,200 bytes.
            bucket_name = self.upload_to_s3_if_required(aws_client, template_path)
            with futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
                results = {executor.submit(self.process_each_test_case, test_case, template_path, aws_client)
                           : test_case for test_case in test_data}
                all_futures.update(results)
                for future in futures.as_completed(all_futures):
                    test_case = all_futures[future]
                    test_case_validations = []
                    try:
                        test_case_validations = future.result()
                        full_output.append({"TestName": test_case["TestName"], "Region": test_case["Region"],
                                            "Validations": test_case_validations})
                    except Exception as exc:
                        # traceback.print_exc()
                        self.logger.error("Test Case -> %s failed with validation as %s", test_case["TestName"],
                                          test_case_validations)
            # Generate Report for Console and Report File.
            ReportGeneration.generate_report(full_output, self.test_file_name, template_path, self.logger)
            # Clean Up for S3 Bucket
            aws_client.delete_bucket(bucket_name)
        else:
            self.logger.error("Validation Failed for Test File with Errors as : \n %s" % (
                json.dumps(test_file_validation_errors, indent=4, default=ReportGeneration.object_to_dictionary)))

    def process_each_test_case(self, test_case, template_path, aws_client):
        test_case_validations = []
        try:
            for process in self.process_flow:
                process_name, process_errors = self.run_process(process, test_case, template_path, aws_client)
                if process_errors:
                    test_case_validations.append(process_errors)
                    if check_for_error(process_name, process_errors):
                        break
        finally:
            process_name, process_errors = self.run_process('cleanup.cleanup.CleanUp', test_case, template_path,
                                                            aws_client)
            if process_errors:
                test_case_validations.append(process_errors)

        return test_case_validations

    def run_process(self, process, test_case, template_path, aws_client):
        process_instance = get_instance_of_process(process)(self.logger, test_case, template_path, aws_client)
        output = process_instance.transform_output(process_instance.process())
        if output:
            return process_instance.process_name, output
        return None, None

    def upload_to_s3_if_required(self, aws_client, template_path):
        output_file = cfn_flip.to_json(read_file(template_path))
        size = sys.getsizeof(output_file)
        if size > 51200:
            self.logger.info("Size of the Template -> %s which is greater than max limit of 51,200 bytes.", size)
            bucket_name = string.ascii_lowercase
            bucket_name = ''.join(random.choice(bucket_name) for i in range(10))
            bucket_name = "sumo-tf-" + bucket_name
            s3_client = aws_client.s3_client_instance("us-east-1")
            aws_client.create_bucket(s3_client, bucket_name)
            time.sleep(30)
            file_url = aws_client.upload_to_bucket(s3_client, bucket_name, template_path)
            self.logger.info("Uploaded template to S3 with URL as %s", file_url)
            return bucket_name
        return None



from common.objects import ProcessOutput, Level, ProcessOutputs

from deploy.deploy import Deploy

from common.cf_utility import Stack


class DeployStack(Deploy):

    def __init__(self, logger, test_case, template_path, aws_client):
        super().__init__(logger, test_case, template_path, aws_client)
        self.parameters = self.test_data["parameters"]
        self.assertions = self.test_data["assertions"]
        self.stack = Stack(test_name=self.test_name, region=self.region, test_data=self.test_data,
                           template_path=template_path, logger=logger, aws_client=aws_client)

    def sub_process_name(self) -> str:
        return "Deploy-Stack"

    """
    Deploy template provided and produces errors and warnings.
    """

    def sub_process(self) -> ProcessOutputs:
        process_outputs = ProcessOutputs()
        status, errors = self.stack.deploy_template()
        if status != "PASS":
            for error in errors:
                error = ProcessOutput(level=Level.ERROR, resource_name=error["Resources"],
                                      message=error["Message"],
                                      stack_name=self.stack.stack_name)
                process_outputs.append(error)
        self.logger.debug("Deploy Stack complete with errors as %s", process_outputs)
        return process_outputs



from common.objects import ProcessOutput, Level, ProcessOutputs
from process import BaseProcess
from common.utils import get_instance_of_process


class Deploy(BaseProcess):
    """
        - Create the parameters to deploy the template
        - Deploy the template along with parameters using deploy with s3 bucket.
    """

    def __init__(self, logger, test_case, template_path, aws_client):
        super().__init__(logger, "Deploy")
        self.test_case = test_case
        self.test_name = test_case["TestName"]
        self.region = test_case["Region"]
        self.test_data = test_case["TestData"]
        self.template_path = template_path
        self.aws_client = aws_client

    # This method should be used by child classes to provide a human readable sub process name.
    def sub_process_name(self) -> str:
        raise NotImplementedError()

    # this method should be used to implement the sub_process process. The method should return ProcessOutputs object.
    def sub_process(self) -> ProcessOutputs:
        raise NotImplementedError()

    process_flow = ['deploy.parameter_check.ParameterCheck', 'deploy.stack_deploy.DeployStack']

    def process(self) -> dict:
        self.logger.info("Started stage -> %s for Test Name -> %s in Region -> %s.", self.process_name,
                         self.test_name, self.region)
        errors = {}
        # First validate the parameters are provided correctly for each test case
        # Then deploy each test case in regions provided using threading.
        for sub_process in self.process_flow:
            validation_errors = ProcessOutputs()
            sub_process_name = "Deploy"
            try:
                class_instance = get_instance_of_process(sub_process)(self.logger, self.test_case, self.template_path,
                                                                      self.aws_client)
                sub_process_name = class_instance.sub_process_name()
                validation_errors = class_instance.sub_process()
            except Exception as exception:
                error = ProcessOutput(level=Level.ERROR, resource_name="Deploy Exception", message=str(exception))
                validation_errors.append(error)
            finally:
                if validation_errors:
                    errors[sub_process_name] = validation_errors
                    break
        self.logger.info("Completed stage -> %s for Test Name -> %s in Region -> %s.", self.process_name,
                         self.test_name, self.region)
        return errors
